/*
 * Licensed Materials - Property of IBM
 *
 * OCO Source Materials
 * 
 * (C) Copyright IBM Corp. 2007, 2010 All Rights Reserved
 *
 * The source code for this program is not published or other-
 * wise divested of its trade secrets, irrespective of what has
 * been deposited with the U.S. Copyright Office.
 */

#ifdef WIN32
  #include <windows.h>
  #include <winioctl.h>
  #include <stdio.h>
#else
  #include <string.h>
  #include <stdio.h>
  #include <stdlib.h>
  #include <stdint.h>
  #include <fcntl.h>
  #include <unistd.h>
  #include <sys/mman.h>
  #include <sys/types.h>
  #include <asm/page.h>
#endif

struct _smHeader {
  char           signature[5]; /* 0 - 4   */
  char           reserved1;    /* 5       */
  unsigned char  version;      /* 6       */
  unsigned char  revision;     /* 7       */
  char           reserved2[8]; /* 8 - 15  */
  unsigned short biosLength;   /* 16 - 17 */
  unsigned       biosAddr;     /* 18 - 21 */
  char           reserved3[10];/* 22 - 27 */
  unsigned short number;       /* 28 - 29 */
  char           reserved4[30];/* 30 - 63 */

} smHeader;

struct _dmiHeader {
  char           signature[5]; /* 0 - 4   */
  char           reserved1;    /* 5       */
  unsigned short biosLength;   /* 6 - 7   */
  unsigned       biosAddr;     /* 8 - 11  */
  unsigned short number;       /* 12 - 13 */
  unsigned char  version;      /* 14      */
  char           reserved2;    /* 15      */

} dmiHeader;

typedef struct _biosHeader {
  char           type;              /* 0   */
  unsigned char  headerLength;      /* 1   */
  unsigned short handle;            /* 2-3 */

} biosHeader;

typedef struct _biosInfoHeader {
  char           type;              /* 0   */
  unsigned char  headerLength;      /* 1   */
  unsigned short handle;            /* 2-3 */
  unsigned char  vendorIndex;       /* 4   */
  unsigned char  versionIndex;      /* 5   */
  unsigned short segmentIndex;      /* 6-7 */
  unsigned char  dateIndex;         /* 8   */
  unsigned char  romSizeIndex;      /* 9   */
  unsigned char  reserved[2];       /* 10+ */
} biosInfoHeader;

typedef struct _sysInfoHeader {
  char           type;              /* 0   */
  unsigned char  headerLength;      /* 1   */
  unsigned short handle;            /* 2-3 */
  unsigned char  manufacturerIndex; /* 4   */
  unsigned char  productIndex;      /* 5   */
  unsigned char  versionIndex;      /* 6   */
  unsigned char  serialIndex;       /* 7   */

} sysInfoHeader;

typedef struct _chassisInfoHeader {
  char           type;              /* 0   */
  unsigned char  headerLength;      /* 1   */
  unsigned short handle;            /* 2-3 */
  unsigned char  manufacturerIndex; /* 4   */
  unsigned char  chassisType;       /* 5   */

} chassisInfoHeader;


typedef struct _procInfoHeader {
  char           type;              /* 0   */
  unsigned char  headerLength;      /* 1   */
  unsigned short handle;            /* 2-3 */
  unsigned char  sockedDesignIndex; /* 4   */
  unsigned char  processorType;     /* 5   */
  unsigned char  processorFamily;   /* 6   */
  unsigned char  manufacturerIndex; /* 7   */
  unsigned char  processorID[8];    /* 8   */
  unsigned char  versionIndex;      /* 16  */
  unsigned char  voltage;           /* 17  */
  unsigned short externalClock;     /* 18  */
  unsigned short maximumSpeed;      /* 20  */
  unsigned short currentSpeed;      /* 22  */
  unsigned char  status;            /* 24  */
  unsigned char  processorUpgrade;  /* 25  */
  unsigned short l1CacheHandle;     /* 26  */
  unsigned short l2CacheHandle;     /* 28  */
  unsigned short l3CacheHandle;     /* 30  */
  unsigned char  serialNumIndex;    /* 32  */
  unsigned char  assetTagIndex;     /* 33  */
  unsigned char  partNumberIndex;   /* 34  */

} procInfoHeader;

typedef struct _memInfoHeader {
  char           type;              /* 0   */
  unsigned char  headerLength;      /* 1   */
  unsigned short handle;            /* 2-3 */
  unsigned short memArrayHandle;    /* 4   */
  unsigned short errorInfoHandle;   /* 6   */
  unsigned short totalWidth;        /* 8   */
  unsigned short dataWidth;         /* 10  */
  unsigned short memorySize;        /* 12  */
  unsigned char  formFactor;        /* 14  */
  unsigned char  deviceSetIndex;    /* 15  */
  unsigned char  devLocatorIndex;   /* 16  */
  unsigned char  bankLocatorIndex;  /* 17  */
  unsigned char  memoryType;        /* 18  */
  unsigned short memTypeDetail;     /* 19  */
  unsigned short memorySpeed;       /* 21  */
  unsigned char  manufacturerIndex; /* 23  */
  unsigned char  serialNumIndex;    /* 24  */
  unsigned char  assetTagIndex;     /* 25  */
  unsigned char  partNumberIndex;   /* 26  */

} memInfoHeader;

void printBiosInfo(biosInfoHeader * biosInfo);
void printSysInfo(sysInfoHeader * sysInfo, int isNewBiosVersion);
void printChassisInfo(sysInfoHeader * sysInfo);
void printProcInfo(procInfoHeader * procInfo);
void printProcVendor(procInfoHeader * procInfo);
void printMemInfo(memInfoHeader * memInfo);
void parseMachineTypeModel(char* productId, char* machType, char* model);
int notEnclosure(chassisInfoHeader * chassisInfo);
char* getBiosString(char* biosHdr, int strNum);
int openPhysicalMemory();
void closePhysicalMemory();
int readPhysicalMemory(void * buffer, unsigned long physAddr, int length);
unsigned long searchPhysicalMemory(char * match, unsigned long startAddr, 
                                   unsigned long endAddr, int stepSize);

#ifdef WIN32
  int isNT();
  int openPhysicalMemoryNT();
  int openPhysicalMemory95();
  int readPhysicalMemoryNT(void * buffer, unsigned long physAddr, int length);
  int readPhysicalMemory95(void * buffer, unsigned long physAddr, int length);
  unsigned long searchPhysicalMemoryNT(char * match, unsigned long startAddr, 
                                       unsigned long endAddr, int stepSize);
  unsigned long searchPhysicalMemory95(char * match, unsigned long startAddr, 
                                       unsigned long endAddr, int stepSize);
  int mapPhysicalMemory(unsigned long * address, unsigned long * length, void ** virtualAddress);
  void unmapPhysicalMemory(void * address);
  int locateNtdllEntryPoints();

  typedef LARGE_INTEGER PHYSICAL_ADDRESS;
  #define PMEM_IOCTL_READ CTL_CODE(0x041274, 0x900, METHOD_BUFFERED, FILE_READ_ACCESS)
  #define PMEM_IOCTL_SEARCH CTL_CODE(0x041274, 0x901, METHOD_BUFFERED, FILE_READ_ACCESS)

  typedef enum _SECTION_INHERIT { ViewShare = 1, ViewUnmap = 2 } SECTION_INHERIT;
  typedef struct { unsigned long addr1; unsigned long addr2; 
                   short length; short step; char data[1]; }  TPMemParams;
  typedef struct _UNICODE_STRING { USHORT Length; USHORT MaximumLength; PWSTR  Buffer; } UNICODE_STRING;
  typedef struct _OBJECT_ATTRIBUTES { ULONG Length; HANDLE RootDirectory; UNICODE_STRING * ObjectName;
                                      ULONG Attributes; PVOID SecurityDescriptor; 
                                      PVOID SecurityQualityOfService; } OBJECT_ATTRIBUTES;

  typedef ULONG (__stdcall *RtlNtStatusToDosError1)(IN LONG Status);
  typedef LONG (__stdcall *NtUnmapViewOfSection1)(IN HANDLE ProcessHandle, IN PVOID BaseAddress);
  typedef VOID (__stdcall *RtlInitUnicodeString1)(IN OUT UNICODE_STRING * DestinationString, IN PCWSTR SourceString);
  typedef LONG (__stdcall *NtOpenSection1)(OUT PHANDLE SectionHandle, IN ACCESS_MASK DesiredAccess,
                                  IN OBJECT_ATTRIBUTES * ObjectAttributes);
  typedef LONG (__stdcall *NtMapViewOfSection1)(IN HANDLE SectionHandle, IN HANDLE ProcessHandle,
                                       IN OUT PVOID * BaseAddress, IN ULONG ZeroBits,
                                       IN ULONG CommitSize, IN OUT PLARGE_INTEGER SectionOffset,
                                       IN OUT PULONG ViewSize, IN SECTION_INHERIT InheritDisposition,
                                       IN ULONG AllocationType, IN ULONG Protect);

  NtOpenSection1 NtOpenSection; 
  NtMapViewOfSection1 NtMapViewOfSection; 
  NtUnmapViewOfSection1 NtUnmapViewOfSection; 
  RtlInitUnicodeString1 RtlInitUnicodeString; 
  RtlNtStatusToDosError1 RtlNtStatusToDosError; 

  static HANDLE deviceHandle = INVALID_HANDLE_VALUE;
  static HANDLE memoryHandle = INVALID_HANDLE_VALUE;
#else
  static int memoryHandle = -1;
#endif


////////////////////////////////////////////////////
//  main(int argc, char * argv[])
////////////////////////////////////////////////////
int main(int argc, char * argv[])
{
    biosHeader * biosHdr;
    char * strBase, * buffer = NULL;
    unsigned int address = 0, biosAddr = 0;
    int i, j, biosLength, biosNumber, reqType = 0, nodeCount = 0, isNewBiosVersion=0;
    unsigned char specMajor, specMinor;

    // Attempt to Open up Physical Memory for reading
    if (!openPhysicalMemory())
    {  printf("Failed - Error Opening Physical Memory\n"); return 0;  }

    // Search through physical memory for any anchor string of _SM_
    if ((address = searchPhysicalMemory("_Sx2xM_", 0xE0000, 0xFFFFF, 16)) != 0)
    {
        if (readPhysicalMemory(&smHeader, address, sizeof(smHeader)))
        {  printf("Failed - Error Reading _SM_ Header\n"); return 0;  }

        biosAddr = smHeader.biosAddr;
        biosLength = smHeader.biosLength;
        biosNumber = smHeader.number;
    }
    // Search through physical memory for any anchor string of _DMI_
    else if ((address = searchPhysicalMemory("_DMI_", 0xE0000, 0xFFFFF, 16)) != 0)
    {
        if (readPhysicalMemory(&dmiHeader, address, sizeof(dmiHeader)))
        {  printf("Failed - Error Reading _DMI_ Header\n"); return 0;  }

        biosAddr = dmiHeader.biosAddr;
        biosLength = dmiHeader.biosLength;
        biosNumber = dmiHeader.number;
        unsigned char tempversion= dmiHeader.version;

        specMajor= (tempversion&240) >> 4; //AND with 240 (xF0) and shift bits to get the upper nibble
        specMinor= tempversion&15;         //AND with 15  (x0F) to get the lower nibble
        if( specMajor > 2 || (specMajor == 2 && specMinor >= 6 && specMinor < 10) || (specMajor == 2 && specMinor >= 60 && specMinor < 100)) 
            isNewBiosVersion = 1;
    }
    // Anchor strings weren't found
    else
    {  printf("Failed - Anchor String Not Found\n"); return 0;  }

    printf("BiosExeVersion=8\n");

    buffer = (char*)malloc(biosLength);

    // Attempt to read the BIOS info once we have found the anchor string
    if (readPhysicalMemory(buffer, biosAddr, biosLength))
    {  free(buffer); printf("Failed - Error Reading BIOS Info\n"); return 0;  }

    biosHdr = (biosHeader*)buffer;

    // If they gave us a SMBIOS Type use that in the request
    if (argc > 1 && atoi(argv[1]) > 0 && atoi(argv[1]) < 10)
        reqType = atoi(argv[1]);

    // Loop through each BIOS record, parsing out the essential pieces of inf
    for(i=0; i<biosNumber; i++)
    {
        // Type 0 is the Base BIOS Info such as BIOS version, data, etc.
        if (biosHdr->type == 0 && reqType == 0)
            printBiosInfo((biosInfoHeader*)biosHdr);
        // Type 1 is the System Information, like UUID, Model, Serial, etc.
        else if (biosHdr->type == 1 && reqType == 0)
            printSysInfo((sysInfoHeader*)biosHdr, isNewBiosVersion);
        // Type 1 is used for Chassis, like UUID, Model, Serial, etc.
        else if (biosHdr->type == 1 && reqType == 3)
            printChassisInfo((sysInfoHeader*)biosHdr);
        // Type 3 is the System Enclosure / Chassis Information
        else if (biosHdr->type == 3 && reqType == 0) {
            if(notEnclosure((chassisInfoHeader*)biosHdr)){
               nodeCount++;
            }
        }
        // Type 4 is the Processor Information
        else if (biosHdr->type == 4 && reqType == 0)
            printProcVendor((procInfoHeader*)biosHdr);
        // Type 4 is the Processor Information
        else if (biosHdr->type == 4 && reqType == 4)
            printProcInfo((procInfoHeader*)biosHdr);
        // Type 6/17 is the Memory Information
        else if (biosHdr->type == 17 && reqType == 6)
            printMemInfo((memInfoHeader*)biosHdr);

        strBase = ((char*)biosHdr) + biosHdr->headerLength;
        // Parse ahead to the next record to see if it is one we care about
        if (strBase[0] == 0x00 && strBase[1] == 0x00)
            biosHdr = (biosHeader*)(strBase+2);
        // Parse ahead to the next record to see if it is one we care about
        else
        {
            for(j=1; *strBase!=0 && j<256; j++)
                strBase += strlen(strBase) + 1;
            biosHdr = (biosHeader*)(strBase+1);
        }
    }

    // Write out if it is a MultiNode or not based on the number of Processors
    if (nodeCount > 1) {
        printf("VMID=0\n");
        printf("MultiNode=true\n");
    }
    else if (reqType == 0)
        printf("MultiNode=false\n");

    free(buffer);

    // Close Physical Memory when we are done
    closePhysicalMemory();

    return 0;
}

////////////////////////////////////////////////////
//  printBiosInfo
////////////////////////////////////////////////////
void printBiosInfo(biosInfoHeader * biosInfo)
{
    int j;
    char * strBase = ((char*)biosInfo) + biosInfo->headerLength;

    // Parse out the BIOS Information that is of use to us to stdout
    for(j=1; *strBase!=0 && j<256; j++)
    {
        if (j == biosInfo->vendorIndex)
            printf("BiosVendor=%s\n", strBase);
        else if (j == biosInfo->versionIndex)
            printf("BiosVersion=%s\n", strBase);
        else if (j == biosInfo->dateIndex)
            printf("BiosDate=%s\n", strBase);
        else if (j == biosInfo->romSizeIndex)
            printf("BiosRomSize=%s\n", strBase);

        strBase += strlen(strBase) + 1;
    }
}

////////////////////////////////////////////////////
//  printSysInfo
////////////////////////////////////////////////////
void printSysInfo(sysInfoHeader * sysInfo, int isNewBiosVersion)
{
    int j;
    char emptyUUID[16];
    char * strBase = ((char*)sysInfo) + sysInfo->headerLength;

    memset(emptyUUID, 0xFF, 16);
    // Parse out the System Information that is of use to us to stdout
    for(j=1; *strBase!=0 && j<256; j++)
    {
        if (j == sysInfo->manufacturerIndex)
            printf("SystemManufacturer=%s\n", strBase);
        else if (j == sysInfo->productIndex)
            printf("SystemProduct=%s\n", strBase);
        else if (j == sysInfo->serialIndex)
            printf("SystemSerialNumber=%s\n", strBase);
        else if (j == sysInfo->versionIndex)
            printf("SystemVersion=%s\n", strBase);

        strBase += strlen(strBase) + 1;
    }

    // The System UUID is a bit tricky, where we want to turn it in to hexadecimal characters
    if (sysInfo->headerLength >= 24)
    {
        // if the spec version is 2.6 or greater we need to reverse the bytes
        if (memcmp(((unsigned char*)sysInfo)+8, emptyUUID, 16) != 0)
        {
            printf("SystemUUID=");
            if (isNewBiosVersion) {
                for(j=11; j>=8; j--)
                    printf("%2.2X", ((unsigned char*)sysInfo)[j]);
                for(j=13; j>=12; j--)
                    printf("%2.2X", ((unsigned char*)sysInfo)[j]);
                for(j=15; j>=14; j--)
                    printf("%2.2X", ((unsigned char*)sysInfo)[j]);
                for(j=16; j<24; j++)
                    printf("%2.2X", ((unsigned char*)sysInfo)[j]);
             }           
             else {
                for(j=8; j<24; j++)
                    printf("%2.2X", ((unsigned char*)sysInfo)[j]);
             }
             printf("\n");			           
        }
    }
}

////////////////////////////////////////////////////
//  printChassisInfo
////////////////////////////////////////////////////
void printChassisInfo(sysInfoHeader * sysInfo)
{
    int j;
    char emptyUUID[16], vendor[512], productId[512], serialNum[512], machType[64], model[512];
    char * strBase = ((char*)sysInfo) + sysInfo->headerLength;

    memset(emptyUUID, 0xFF, 16);
    memset(vendor, 0x00, 512);
    memset(productId, 0x00, 512);
    memset(serialNum, 0x00, 512);
    // Parse out the System Information that is of use to us to stdout
    for(j=1; *strBase!=0 && j<256; j++)
    {
        if (j == sysInfo->manufacturerIndex)
            sprintf(vendor, "%s", strBase);
        else if (j == sysInfo->productIndex)
            sprintf(productId, "%s", strBase);
        else if (j == sysInfo->serialIndex)
            sprintf(serialNum, "%s", strBase);

        strBase += strlen(strBase) + 1;
    }

    parseMachineTypeModel(productId, machType, model);

    printf("Chassis.0.Name=%s,%s-%s-%s\n", vendor, machType, model, serialNum);
    printf("Chassis.0.Tag=%s %s%s %s\n", vendor, machType, model, serialNum);
    printf("Chassis.0.Manufacturer=%s\n", vendor);
    printf("Chassis.0.Model=%s\n", model);
    printf("Chassis.0.VendorType=%s\n", machType);
    printf("Chassis.0.SerialNumber=%s\n", serialNum);
    printf("Chassis.0.PackageType=3\n");

    // The System UUID is a bit tricky, where we want to turn it in to hexadecimal characters
    if (sysInfo->headerLength >= 24)
    {
        if (memcmp(((unsigned char*)sysInfo)+8, emptyUUID, 16) != 0)
        {
            printf("Chassis.0.PackageGUID=");
            for(j=8; j<12; j++)
                printf("%2.2X", ((unsigned char*)sysInfo)[j]);
            printf("-");
            for(j=12; j<14; j++)
                printf("%2.2X", ((unsigned char*)sysInfo)[j]);
            printf("-");
            for(j=14; j<16; j++)
                printf("%2.2X", ((unsigned char*)sysInfo)[j]);
            printf("-");
            for(j=16; j<18; j++)
                printf("%2.2X", ((unsigned char*)sysInfo)[j]);
            printf("-");
            for(j=18; j<24; j++)
                printf("%2.2X", ((unsigned char*)sysInfo)[j]);
            printf("\n");
        }
    }
}

////////////////////////////////////////////////////
//  printProcInfo
////////////////////////////////////////////////////
void printProcInfo(procInfoHeader * procInfo)
{
    static int procCnt = 0;
    int width = 32, curSpeed = 0;
    char * strBase = ((char*)procInfo) + procInfo->headerLength;

    if(!((procInfo->status & 0x40) && ((procInfo->status & 0x07) == 0x1)))
        return;

    curSpeed = procInfo->currentSpeed == 0 ? procInfo->maximumSpeed : procInfo->currentSpeed;

    if (procInfo->versionIndex > 0)
        printf("Processor.%d.Name=%s\n", procCnt, getBiosString(strBase,procInfo->versionIndex));
    printf("Processor.%d.DeviceID=CPU%d\n", procCnt, procCnt+1);
    if (procInfo->processorFamily > 0 && procInfo->processorFamily < 250)
        printf("Processor.%d.Family=%d\n", procCnt, procInfo->processorFamily);
    printf("Processor.%d.AddressWidth=%d\n", procCnt, width);
    printf("Processor.%d.DataWidth=%d\n", procCnt, width);
    printf("Processor.%d.CurrentClockSpeed=%d\n", procCnt, curSpeed);
    printf("Processor.%d.MaxClockSpeed=%d\n", procCnt, procInfo->maximumSpeed);
    printf("Processor.%d.NumberOfEnabledCores=1\n", procCnt);
    printf("Processor.%d.HealthState=1\n", procCnt);
    printf("Processor.%d.OperatingState=8\n", procCnt);
    if (procInfo->manufacturerIndex > 0)
        printf("Processor.%d.Manufacturer=%s\n", procCnt, getBiosString(strBase,procInfo->manufacturerIndex));

    procCnt++;
}

////////////////////////////////////////////////////
//  printProcVendor
////////////////////////////////////////////////////
void printProcVendor(procInfoHeader * procInfo)
{
    static int procFound = 0;
    if (procFound)
        return;

    char * strBase = ((char*)procInfo) + procInfo->headerLength;
    if(!((procInfo->status & 0x40) && ((procInfo->status & 0x07) == 0x1)))
        return;

    if (procInfo->manufacturerIndex > 0) {
        printf("ProcVendor=%s\n", getBiosString(strBase,procInfo->manufacturerIndex));
        procFound = 1;
    }
}

const unsigned short FormFactorMap[] = { 1, 0, 7, 2, 1, 3, 4, 6, 8, 9, 1, 11, 12, 13 };
const unsigned short MemoryTypeMap[] = { 1, 0, 2, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20 };

////////////////////////////////////////////////////
//  printMemInfo
////////////////////////////////////////////////////
void printMemInfo(memInfoHeader * memInfo)
{
    static int memCnt = 0;
    uint64_t capacity = 0;
    char * strBase = ((char*)memInfo) + memInfo->headerLength;

    if(memInfo->memorySize == 0 || memInfo->memorySize == 0xFFFF)
        return;
    
    capacity = memInfo->memorySize & 0x7FFF;
    capacity *= 1024;
    if ((memInfo->memorySize & 0x8000) != 0x8000)
        capacity *= 1024;

    if (memInfo->devLocatorIndex > 0)
        printf("PhysicalMemory.%d.Name=%s\n", memCnt, getBiosString(strBase,memInfo->devLocatorIndex));
    printf("PhysicalMemory.%d.Tag=<Chassis.0.Tag>-P1-C%d\n", memCnt, memCnt+1);
    if (memInfo->bankLocatorIndex > 0)
        printf("PhysicalMemory.%d.BankLabel=%s\n", memCnt, getBiosString(strBase,memInfo->bankLocatorIndex));
    printf("PhysicalMemory.%d.DataWidth=%d\n", memCnt, memInfo->dataWidth);
    printf("PhysicalMemory.%d.TotalWidth=%d\n", memCnt, memInfo->totalWidth);
    if (capacity > 0)
        printf("PhysicalMemory.%d.Capacity=%lld\n", memCnt, capacity);
    if (memInfo->formFactor >= 1 && memInfo->formFactor <=14)
        printf("PhysicalMemory.%d.FormFactor=%d\n", memCnt, FormFactorMap[memInfo->formFactor-1]);
    if (memInfo->memoryType >= 1 && memInfo->memoryType <=18)
        printf("PhysicalMemory.%d.MemoryType=%d\n", memCnt, MemoryTypeMap[memInfo->memoryType-1]);
    if (memInfo->headerLength >= 26)
    {
        if (memInfo->memorySpeed > 0)
            printf("PhysicalMemory.%d.Speed=%d\n", memCnt, memInfo->memorySpeed);
        if (memInfo->manufacturerIndex > 0)
            printf("PhysicalMemory.%d.Manufacturer=%s\n", memCnt, getBiosString(strBase,memInfo->manufacturerIndex));
        if (memInfo->serialNumIndex > 0)
            printf("PhysicalMemory.%d.SerialNumber=%s\n", memCnt, getBiosString(strBase,memInfo->serialNumIndex));
        if (memInfo->partNumberIndex > 0)
            printf("PhysicalMemory.%d.PartNumber=%s\n", memCnt, getBiosString(strBase,memInfo->partNumberIndex));
    }

    memCnt++;
}

////////////////////////////////////////////////////
//  parseMachineTypeModel
////////////////////////////////////////////////////
void parseMachineTypeModel(char* productId, char* machType, char* model)
{
    char* openBracket, *closeBracket;
    memset(machType, 0x00, 64);
    memset(model, 0x00, 512);

    openBracket = strchr(productId, '[');
    if (openBracket != NULL)
        productId = openBracket + 1;
    closeBracket = strchr(productId, ']');
    if (closeBracket != NULL)
        closeBracket[0] = 0x00;

    if (strlen(productId) == 7)
    {
        memcpy(machType, productId, 4);
        memcpy(model, productId+4, 3);
    }
    else
        sprintf(model, "%s", productId); 
}

int notEnclosure(chassisInfoHeader * chassisInfo) {

    // Bit 7 Chassis lock is present if 1.Bits 6:0 Enumeration value
    unsigned char chassisType = chassisInfo->chassisType;
    unsigned char thetype = chassisType&127;         //AND with 127 (0x7F) to get the first 7 bits

    if(thetype == 29 || thetype == 18) { //0x1D is an enclosure, //0x12 is an expansion
       return 0;
    }
    return 1;
}

////////////////////////////////////////////////////
//  getBiosString
////////////////////////////////////////////////////
char* getBiosString(char* biosHdr, int strNum)
{
    int j = 1;
    char * strBase = biosHdr;

    while (j < strNum)
    {
       while((*strBase) != 0x00)
         strBase++;
       strBase++;
       j++;
    }
    return strBase;
}

#ifndef WIN32

////////////////////////////////////////////////////
//  openPhysicalMemory
////////////////////////////////////////////////////
int openPhysicalMemory()
{
    memoryHandle = open("/dev/mem",O_RDONLY);

    return memoryHandle >= 0;
}

////////////////////////////////////////////////////
//  closePhysicalMemory
////////////////////////////////////////////////////
void closePhysicalMemory()
{
    if (memoryHandle >= 0)
        close(memoryHandle);
}

////////////////////////////////////////////////////
//  readPhysicalMemory
////////////////////////////////////////////////////
int readPhysicalMemory(void * buffer, unsigned long physAddr, int length)
{
	char* virtualAddr;
	unsigned int pageAddr;
	int pagesize = getpagesize();

	pageAddr = physAddr & ~(getpagesize()-1);
	virtualAddr = (char*)mmap(0, getpagesize(), PROT_READ, MAP_SHARED, memoryHandle, pageAddr);

    // If we have hit the point to switch pages, unmap the old page, and map a new page
    // we could be starting in the middle of a page
	if (pageAddr + getpagesize() < physAddr+length)
	{
		int copylength = (pageAddr + getpagesize()) - physAddr ; // could be the last part of the first page
		int bytescopied=0;
		while (bytescopied < length) {
			memcpy(((char*)buffer)+bytescopied , virtualAddr + ((physAddr+bytescopied) - pageAddr), copylength);
			bytescopied +=copylength;
			munmap(virtualAddr, getpagesize());
			pageAddr = (physAddr+bytescopied ) & ~(getpagesize()-1);
			virtualAddr = (char*)mmap(0, getpagesize(), PROT_READ, MAP_SHARED, memoryHandle, pageAddr);

			// copy a page at a time, we already copied the first page, now every page until the last
			// one will be a full page size.
			copylength = pagesize;
			if((copylength + bytescopied ) > length) {
				copylength = length - bytescopied;
			}
		}
	}
    // If we are on the correct page, just copy over the data at the given address
	else
		memcpy(buffer, virtualAddr + (physAddr - pageAddr), length);

    munmap(virtualAddr, getpagesize());
	return 0;
}

////////////////////////////////////////////////////
//  searchPhysicalMemory
////////////////////////////////////////////////////
unsigned long searchPhysicalMemory(char * match, unsigned long startAddr, 
                                   unsigned long endAddr, int stepSize)
{
	int rc, matchLength;
	unsigned int i, pageAddr;
	char * buffer, * virtualAddr;

	matchLength = strlen(match);
	buffer = (char*)malloc(matchLength+1);
	buffer[matchLength] = 0;

	pageAddr = startAddr & ~(getpagesize()-1);
	virtualAddr = (char *)mmap(0, getpagesize(), PROT_READ, MAP_SHARED, memoryHandle, pageAddr);
	if ((int)virtualAddr == -1) { 
		free(buffer); 
		return 0;
	}

    // Iterate through the entire section, skipping by step sizes
	for (i=startAddr; i<endAddr; i+=stepSize)
	{
    // If we have hit the point to switch pages, unmap the old page, and map a new page
		if (i-pageAddr >= getpagesize())
		{
			munmap(virtualAddr, getpagesize());
			pageAddr = i & ~(getpagesize()-1);
			virtualAddr = (char *)mmap(0, getpagesize(), PROT_READ, MAP_SHARED, memoryHandle, pageAddr);
			if ((int)virtualAddr == -1) { 
				free(buffer); 
				return 0;
			}
		}
		memcpy(buffer, virtualAddr + (i-pageAddr), matchLength);
		rc = strncmp(match, buffer, matchLength);
		if (rc == 0)
        {  free(buffer); munmap(virtualAddr, getpagesize()); return i;  }
	}

	munmap(virtualAddr, getpagesize());
	free(buffer);
	return(0);
}


#else

////////////////////////////////////////////////////
//  isNT
////////////////////////////////////////////////////
int isNT()
{
    static int osNT = 0;
    static int retrieved = 0;

    // We only need to retrieve whether it is NT once since it isn't going to change
    if (!retrieved)
    {
         OSVERSIONINFO osver;
         osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

         // Get the Version Information to determine whether the platform is NT-like or not
         GetVersionEx(&osver);
         osNT = osver.dwPlatformId == VER_PLATFORM_WIN32_NT;
    }

    return osNT;
}

////////////////////////////////////////////////////
//  openPhysicalMemory
////////////////////////////////////////////////////
int openPhysicalMemory()
{
    if (isNT())
        return openPhysicalMemoryNT();
    else
        return openPhysicalMemory95();
}

////////////////////////////////////////////////////
//  openPhysicalMemoryNT
////////////////////////////////////////////////////
int openPhysicalMemoryNT()
{
    LONG status;
    UNICODE_STRING physmemString;
    OBJECT_ATTRIBUTES attributes;

    // If we can't locate the entry points, we might as well fail right away
    if (!locateNtdllEntryPoints())
        return 0;

    // We need to create a Unicode (16 byte) representation of the section name
    RtlInitUnicodeString(&physmemString, L"\\device\\physicalmemory");

    (&attributes)->Length = sizeof( OBJECT_ATTRIBUTES );
    (&attributes)->RootDirectory = NULL;
    (&attributes)->Attributes = 0x00000040L;
    (&attributes)->ObjectName = &physmemString;
    (&attributes)->SecurityDescriptor = NULL;
    (&attributes)->SecurityQualityOfService = NULL;

    // Attempt to open up the section representing physical memory
    status = NtOpenSection(&memoryHandle, SECTION_MAP_READ, &attributes);

    // If we couldn't open the section, no reason to continue, so just fail
    if(status < 0)
        return 0;

    // Open up the PMEM file after we have successfully found the correct section
    deviceHandle = CreateFile("\\\\.\\PMEM", GENERIC_READ, FILE_SHARE_READ, NULL, 
                             OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);

    return deviceHandle != INVALID_HANDLE_VALUE;
}

////////////////////////////////////////////////////
//  openPhysicalMemory
////////////////////////////////////////////////////
int openPhysicalMemory95()
{
   deviceHandle = CreateFile("\\\\.\\PMEM.VXD", 0, 0, 0, CREATE_NEW, FILE_FLAG_DELETE_ON_CLOSE, 0);

   return deviceHandle != INVALID_HANDLE_VALUE;
}

////////////////////////////////////////////////////
//  closePhysicalMemory
////////////////////////////////////////////////////
void closePhysicalMemory()
{
    if (deviceHandle != INVALID_HANDLE_VALUE)
        CloseHandle(deviceHandle);
    if (memoryHandle != INVALID_HANDLE_VALUE)
        CloseHandle(memoryHandle);
}


////////////////////////////////////////////////////
//  readPhysicalMemory
////////////////////////////////////////////////////
int readPhysicalMemory(void * buffer, unsigned long physAddr, int length)
{
    if (isNT())
        return readPhysicalMemoryNT(buffer, physAddr, length);
    else
        return readPhysicalMemory95(buffer, physAddr, length);
}


////////////////////////////////////////////////////
//  readPhysicalMemoryNT
////////////////////////////////////////////////////
int readPhysicalMemoryNT(void * buffer, unsigned long physAddr, int length)
{
    char * buf, * seg;
    unsigned long addr = physAddr;
    unsigned long len = (physAddr % 0x1000) + length;

    // If the address is outside of the first MB, use the I/O Device search method
    if ((physAddr + length) >= 0x100000)
        return readPhysicalMemory95(buffer, physAddr, length);

    // Try to map the memory first that we want to search through
    if (!mapPhysicalMemory(&addr, &len, (void**)&seg))
        return GetLastError();

    buf = seg + (physAddr-addr);
    memcpy(buffer, buf, length);

    unmapPhysicalMemory(seg);
    return 0;
}

////////////////////////////////////////////////////
//  readPhysicalMemory95
////////////////////////////////////////////////////
int readPhysicalMemory95(void * buffer, unsigned long physAddr, int length)
{
    BOOL success;
    TPMemParams params;
    unsigned long outBufferLen;
    params.addr1 = physAddr;
    params.length = (short)length;

    // Use the I/O Device Control to read through physical memory
    success = DeviceIoControl(deviceHandle, PMEM_IOCTL_READ, &params, 
                              sizeof(params), buffer, length, &outBufferLen, NULL);

    return success ? 0 : GetLastError();
}

////////////////////////////////////////////////////
//  searchPhysicalMemory
////////////////////////////////////////////////////
unsigned long searchPhysicalMemory
(char * match, unsigned long startAddr, unsigned long endAddr, int stepSize)
{
    if (isNT())
        return searchPhysicalMemoryNT(match, startAddr, endAddr, stepSize);
    else
        return searchPhysicalMemory95(match, startAddr, endAddr, stepSize);
}

////////////////////////////////////////////////////
//  searchPhysicalMemoryNT
////////////////////////////////////////////////////
unsigned long searchPhysicalMemoryNT
(char * match, unsigned long startAddr, unsigned long endAddr, int stepSize)
{
    char * buf, * seg;
    int i, matchLen, matchFound;
    unsigned long length, physAddr;

    // If we weren't given a valid address, we might as well fail right away
    if (startAddr >= endAddr)
        return 0;

    // If the address is outside of the first MB, use the I/O Device search method
    if (endAddr >= 0x100000)
        return searchPhysicalMemory95(match, startAddr, endAddr, stepSize);

    physAddr = startAddr;
    matchLen = strlen(match);
    length = (startAddr % 0x1000) + (endAddr - startAddr);

    // Try to map the memory first that we want to search through
    if (!mapPhysicalMemory(&physAddr, &length, (void**)&seg))
        return 0;

    buf = seg + (startAddr-physAddr);
    matchFound = 0;
    length = endAddr - startAddr;

    // Loop through the memory provided, searching for a matching string
    for (i=0; (i+matchLen) <= (int)(length+1); i+=stepSize)
        if (memcmp(buf+i, match, matchLen) == 0)
        {  matchFound = 1; break; }

    // Unmap the memory when we are done searching
    unmapPhysicalMemory(seg);

    return matchFound ? (startAddr+i) : 0;
}

////////////////////////////////////////////////////
//  searchPhysicalMemory95
////////////////////////////////////////////////////
unsigned long searchPhysicalMemory95
(char * match, unsigned long startAddr, unsigned long endAddr, int stepSize)
{
    BOOL success;
    TPMemParams *pParams;
    unsigned long retval;
    unsigned long outBufferLen;

    pParams = (TPMemParams*)malloc(sizeof(TPMemParams) + strlen(match));
    pParams->addr1 = startAddr;
    pParams->addr2 = endAddr;
    pParams->length = (short)strlen(match);
    pParams->step = (short)stepSize;
    strcpy(pParams->data, match);

    // Use the I/O Device Control to search through physical memory for the given value
    success = DeviceIoControl(deviceHandle, PMEM_IOCTL_SEARCH, pParams, sizeof(TPMemParams) + 
                              strlen(match), &retval, sizeof(unsigned long), &outBufferLen, NULL);

    free(pParams);

    return success ? retval : 0;
}

////////////////////////////////////////////////////
//  mapPhysicalMemory
////////////////////////////////////////////////////
int mapPhysicalMemory(unsigned long * address, unsigned long * length, void ** virtualAddress)
{
    LONG status;
    PHYSICAL_ADDRESS viewBase;

    *virtualAddress = 0;
    viewBase.HighPart = 0;
    viewBase.LowPart = (*address);

    // Map the section of memory that is provided to us into the address provided
    status = NtMapViewOfSection(memoryHandle, (HANDLE)-1, (PVOID*)virtualAddress, 0L,
                                *length, &viewBase, length, ViewShare, 0, PAGE_READONLY);

    // If there is an error, return that the mapping failed
    if (status < 0)
       return 0;

    *address = viewBase.LowPart;
    return 1;
}

////////////////////////////////////////////////////
//  unmapPhysicalMemory
////////////////////////////////////////////////////
void unmapPhysicalMemory(void * address)
{
    // Unmap the section of memory that is provided to us
    LONG status = NtUnmapViewOfSection((HANDLE)-1, address);

    if (status < 0)
        return;
}

////////////////////////////////////////////////////
//  locateNtdllEntryPoints
////////////////////////////////////////////////////
int locateNtdllEntryPoints()
{
    // Go through the list of Functions to call and load the entry points for each one
   if(!(RtlInitUnicodeString = (RtlInitUnicodeString1)
        GetProcAddress(GetModuleHandle("ntdll.dll"),"RtlInitUnicodeString")))
       return 0;
   if(!(NtUnmapViewOfSection = (NtUnmapViewOfSection1)
        GetProcAddress(GetModuleHandle("ntdll.dll"),"NtUnmapViewOfSection")))
       return 0;
   if(!(NtOpenSection = (NtOpenSection1)GetProcAddress
        (GetModuleHandle("ntdll.dll"),"NtOpenSection")))
       return 0;
   if(!(NtMapViewOfSection = (NtMapViewOfSection1)
        GetProcAddress(GetModuleHandle("ntdll.dll"),"NtMapViewOfSection")))
       return 0;
   if(!(RtlNtStatusToDosError = (RtlNtStatusToDosError1)
        GetProcAddress(GetModuleHandle("ntdll.dll"),"RtlNtStatusToDosError")))
       return 0;

   return 1;
}

#endif
